/* mbed EventFramework Library 
 * Copyright (c) 2012 raulMrello
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to deal
 * in the Software without restriction, including without limitation the rights
 * to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 * copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in
 * all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
 * THE SOFTWARE.
 */

#include "mbed.h"
#include "PCA9685.h"
#include "PCA9685_ServoDrv.h"

#define SUCCESS     0
const uint8_t INIT_CONFIG[]     = {PCA9685_MODE1, PCA9685_MODE1_AI, PCA9685_MODE2_OUTDRV};
const uint8_t SLEEP_MODE[]      = {PCA9685_MODE1, PCA9685_MODE1_SLEEP};
static uint8_t CHANGE_FREQ[]    = {PCA9685_PRE_SCALE, 0};
  

/** Creates an instance of a driver
 * @parm i2c Driver reference
 * @param devAddr SlaveAddress of the i2c device
 * @param oscHz Oscilator frequency in Hz (default=25000000)
 * @param pwmHz PWM frequency in Hz (default=1Khz)
 */
PCA9685_ServoDrv::PCA9685_ServoDrv(I2C *i2c, uint8_t devAddr, uint32_t oscHz, uint32_t pwmHz){
    uint8_t i;
    this->devAddr = devAddr|PCA9685_FIXED_ADDRESS;
    this->oscHz = oscHz;
    this->pwmHz = pwmHz;
    this->i2c = i2c;
    for(i=0;i<16;i++){
        this->minRange[i] = 0;
        this->maxRange[i] = 4095;
        this->dutyValue[i] = 0;
    }
    // clears errors
    this->errcodes = 0;
    // check if device is present
    if (this->i2c->write(this->devAddr, NULL, 0) != SUCCESS){
        errcodes = -4;
        return;
    }
    // initialize driver: enable auto-increment, normal mode and invert output (no driver required)
    if(this->i2c->write(this->devAddr, (const char*)INIT_CONFIG, sizeof(INIT_CONFIG)) != SUCCESS){
        errcodes = -3;        
        return;
    }
    
	
    // set sleep mode to change frequency
    if(this->i2c->write(this->devAddr, (const char*)SLEEP_MODE, sizeof(SLEEP_MODE)) != SUCCESS){
        errcodes = -3;        
        return;
    }
	    
    // change pwm frequency
    CHANGE_FREQ[1] = PCA9685_PWM_FREQ(this->oscHz, this->pwmHz);
   if(this->i2c->write(this->devAddr, (const char*)CHANGE_FREQ, sizeof(CHANGE_FREQ)) != SUCCESS){
        errcodes = -3;        
        return;
    }
 
    // set normal mode
    // initialize driver: enable auto-increment, normal mode and invert output (no driver required)
    if(this->i2c->write(this->devAddr, (const char*)INIT_CONFIG, sizeof(INIT_CONFIG)) != SUCCESS){
        errcodes = -3;        
        return;
    }
 	// delay after exit stop mode
	for(i=0;i<250;i++){}
   
    // set duty values 
    if(SetAllDuty(this->dutyValue, 16) != SUCCESS){
        errcodes = -3;        
        return;
    }
 }

    
/** Sets the range of the PWM duty for a servo
 * @param servoId Servo id (0 to 15)
 * @param minDuty Minimum duty value (default = 0) 
 * @param maxDuty Maximum duty value (default = 4095)
 * @return error code <= 0
 */
int8_t PCA9685_ServoDrv::SetServoRange(uint8_t servoId, uint16_t minDuty, uint16_t maxDuty){
    // check params to avoid malfunctions
    if(servoId >=16 || minDuty >= maxDuty || maxDuty >= 4096)
        return -1;
    this->minRange[servoId] = minDuty;
    this->maxRange[servoId] = maxDuty;
    return 0;
}

  
/** Gets the range of the PWM duty for a servo. Read the internal property of the class
 * @param servoId Servo id (0 to 15)
 * @param minDuty Pointer to get min range 
 * @param maxDuty Pointer to get max range
 * @return error code <= 0
 */
int8_t PCA9685_ServoDrv::GetServoRange(uint8_t servoId, uint16_t *minDuty, uint16_t *maxDuty){
    // check params to avoid malfunctions
    if(servoId >=16 || minDuty == 0 || maxDuty == 0)
        return -1;
    *minDuty = this->minRange[servoId];
    *maxDuty = this->maxRange[servoId];
    return 0;
}
 
  
/** Sets  the PWM duty for a servo
 * @param servoId Servo id (0 to 15)
 * @param duty  Duty value (0 to 1000) 
 * @return error code <= 0
 */
int8_t PCA9685_ServoDrv::SetServoDuty(uint8_t servoId, uint16_t duty){
    uint8_t buffer[5];
    // check params to avoid malfunctions
    if(servoId >=16 || duty >= 4096)
        return -1;
    // check duty out of range
    if(duty < this->minRange[servoId] || duty > this->maxRange[servoId])
        return -3;
    buffer[0] = PCA9685_SERVO0_ON_L+(servoId*4);
    buffer[1] = 0; //ON_L always 0
    buffer[2] = 0; //ON_H always 0
    buffer[3] = (uint8_t)(duty & 0x00ff);
    buffer[4] = (uint8_t)((duty>>8) & 0x00ff);
    // write buffer and check errors
    if(this->i2c->write(this->devAddr, (const char*)buffer, 5) != SUCCESS)
        return -2;
    // if success, update values
    this->dutyValue[servoId] = duty;
    return 0;
}
   
/** Gets  the PWM duty from a servo
 * @param servoId Servo id (0 to 15)
 * @param duty  Pointer to get its duty value 
 * @return error code <= 0
 */
int8_t PCA9685_ServoDrv::GetServoDuty(uint8_t servoId, uint16_t *duty){
    // check params to avoid malfunctions
    if(servoId >=16 || duty == 0)
        return -1;
    *duty = this->dutyValue[servoId];
    return 0;
}

   
/** Gets  the PWM duty from a servo
 * @param servoValues Pointer to the list of duty values
 * @param numServos Number of servo values in the list
 * @returns error code <= 0
 */
int8_t PCA9685_ServoDrv::GetAllDuty(uint16_t *servoValues, uint8_t numServos){
    char addr = PCA9685_SERVO0_ON_L;
    char buffer[96];
	uint8_t i;
    // check params to avoid malfunctions
    if(servoValues == 0)
        return -1;
    // set read pointer at position 0
    if(this->i2c->write(this->devAddr, &addr, 1) != SUCCESS)
        return -2;
    if(this->i2c->read(this->devAddr, buffer, 4*numServos) != SUCCESS)
        return -2;
	for(i=0;i<numServos;i++){
		uint16_t value = buffer[(i*4)+3];
		value = ((value << 8) & 0xff00) + buffer[(i*4)+2];
		this->dutyValue[i] = value;
		*(servoValues+i) = value;
	}
    return 0;
}

    
/** Update all servo duty at one shot
 * @param servoValues Pointer to the list of duty values
 * @param numServos Number of servo values in the list
 * @returns error code <= 0
 */
int8_t PCA9685_ServoDrv::SetAllDuty(uint16_t *servoValues, uint8_t numServos){
    uint8_t i,j;
    uint8_t buffer[1+(16*4)];
    // check params to avoid malfunctions
    if(numServos > 16 || servoValues == 0)
        return -1;    
    // check duties out of range
//     for(i=0;i<numServos;i++){
//         if(*(servoValues+i) < this->minRange[i] || *(servoValues+i) > this->maxRange[i])
//             return -3;
//     }  
    // setup buffer address
    j=0;
    buffer[j++] = PCA9685_SERVO0_ON_L;
    // setup servo values
    for(i=0;i<numServos;i++){
        buffer[j++] = 0; //ON_L always 0
        buffer[j++] = 0; //ON_H always 0
        buffer[j++] = (uint8_t)(*(servoValues+i) & 0x00ff);
        buffer[j++] = (uint8_t)((*(servoValues+i)>>8) & 0x00ff);
    }
    // reset unused servos
    for(i=numServos;i<16;i++){
        buffer[j++] = 0;
        buffer[j++] = 0;
        buffer[j++] = 0;
        buffer[j++] = 0;
    }
    // write buffer and check errors
    if(this->i2c->write(this->devAddr, (const char*)buffer, j) != SUCCESS)
        return -2;
    // if success, update values
    for(i=0;i<numServos;i++){
        this->dutyValue[i] = *(servoValues+i);
    }      
    return 0;
}

       
/** Get driver content, by register
 * @returns error code <= 0
 */
int8_t PCA9685_ServoDrv::GetDriverContent(void){
    char addr = 0;
    char buffer[96];
    buffer[0] = 0;
    // set read pointer at position 0
    if(this->i2c->write(this->devAddr, &addr, 1) != SUCCESS)
        return -2;
    if(this->i2c->read(this->devAddr, buffer, 70) != SUCCESS)
        return -2;
    // set read pointer at position 0xfa
    addr = 0xfa;
    if(this->i2c->write(this->devAddr, &addr, 1) != SUCCESS)
        return -2;
    if(this->i2c->read(this->devAddr, buffer+70, 5) != SUCCESS)
        return -2;
    return 0;
}

    
/** Set PWM frequency in Hz
 * @param freq Frequency in Hz
 * @returns error code <= 0
 */
int8_t PCA9685_ServoDrv::SetOutputFrequency(uint32_t freq){
	uint8_t i;
	if(!freq)
		return  -1;
	this->pwmHz = freq;
	
    // set sleep mode to change frequency
    if(this->i2c->write(this->devAddr, (const char*)SLEEP_MODE, sizeof(SLEEP_MODE)) != SUCCESS){
        return -3;
    }
	    
    // change pwm frequency
    CHANGE_FREQ[1] = PCA9685_PWM_FREQ(this->oscHz, this->pwmHz);
   if(this->i2c->write(this->devAddr, (const char*)CHANGE_FREQ, sizeof(CHANGE_FREQ)) != SUCCESS){
        return -3;
    }
 
    // set normal mode
    // initialize driver: enable auto-increment, normal mode and invert output (no driver required)
    if(this->i2c->write(this->devAddr, (const char*)INIT_CONFIG, sizeof(INIT_CONFIG)) != SUCCESS){
        return -3;
    }
 	// delay after exit stop mode
	for(i=0;i<250;i++){}
			   
    // set duty values 
    if(SetAllDuty(this->dutyValue, 16) != SUCCESS){
        return -3;
    }
	
	return 0;
	 
}
