/* mbed EventFramework Library
 * Copyright (c) 2012-2016 raulMrello
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to deal
 * in the Software without restriction, including without limitation the rights
 * to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 * copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in
 * all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
 * THE SOFTWARE.
 */

#ifndef __PCA9685_SERVODRV_H
#define __PCA9685_SERVODRV_H

#include "mbed.h"

/** PCA9685_ServoDrv class
 *
 * Servo Driver to control up to 16 servo-motors via I2C bus. It can raise these
 * error codes:
 * -1 Invalid function arguments
 * -2 Failed on device write
 * -3 Duty out of range
 * -4 I2C device not detected
 */
class PCA9685_ServoDrv{
    public:
    
    int32_t errcodes;
    
    /** Creates an instance of a driver (I2C)
     * @parm i2c Driver reference
     * @param devAddr SlaveAddress of the i2c device
     * @param oscHz Oscilator frequency in Hz (default=25000000)
     * @param pwmHz PWM frequency in Hz (default=1Khz)
     */
    PCA9685_ServoDrv(I2C *i2c, uint8_t devAddr, uint32_t oscHz=25000000, uint32_t pwmHz=50);

    /** Destructor
     */
    ~PCA9685_ServoDrv();
    
    
    /** Sets the range of the PWM duty for a servo
     * @param servoId Servo id (0 to 15)
     * @param minDuty Minimum duty value (default = 0) 
     * @param maxDuty Maximum duty value (default = 4095)
     * @return error code <= 0
     */
    int8_t SetServoRange(uint8_t servoId, uint16_t minDuty=0, uint16_t maxDuty=4095);
   
    /** Gets the range of the PWM duty for a servo. Read the internal property of the class
     * @param servoId Servo id (0 to 15)
     * @param minDuty Pointer to get min range 
     * @param maxDuty Pointer to get max range
     * @return error code <= 0
     */
    int8_t GetServoRange(uint8_t servoId, uint16_t *minDuty, uint16_t *maxDuty);
   
    /** Sets  the PWM duty for a servo
     * @param servoId Servo id (0 to 15)
     * @param duty  Duty value (0 to 1000) 
     * @return error code <= 0
     */
    int8_t SetServoDuty(uint8_t servoId, uint16_t duty);
   
    /** Gets  the PWM duty from a servo
     * @param servoId Servo id (0 to 15)
     * @param duty  Pointer to get its duty value 
     * @return error code <= 0
     */
    int8_t GetServoDuty(uint8_t servoId, uint16_t *duty);
    
    /** Get all servo duty at one shot
     * @param servoValues Pointer to the list of duty values
     * @param numServos Number of servo values in the list
     * @returns error code <= 0
     */
    int8_t GetAllDuty(uint16_t *servoValues, uint8_t numServos=16);
    
    /** Update all servo duty at one shot
     * @param servoValues Pointer to the list of duty values
     * @param numServos Number of servo values in the list
     * @returns error code <= 0
     */
    int8_t SetAllDuty(uint16_t *servoValues, uint8_t numServos=16);
    
    /** Get driver content, by register
     * @returns error code <= 0
     */
    int8_t GetDriverContent(void);
    
    /** Set PWM frequency in Hz
     * @param freq Frequency in Hz
     * @returns error code <= 0
     */
    int8_t SetOutputFrequency(uint32_t freq);
    
    private:

    uint16_t minRange[16];
    uint16_t maxRange[16];
    uint16_t dutyValue[16];
    uint8_t devAddr;
    uint32_t oscHz;
    uint32_t pwmHz;
    I2C *i2c; 
};

#endif
